import pyamg
import tensorflow as tf
import numpy as np
import copy
import itertools
from collections.abc import Iterator

from .cholesky import poisson_RHS

# def poisson_RHS(F, boundaries = None, h = None, rho = None): #Currently unused method to handle Poisson systems where BC mesh nodes are included in the LHS vector
#     boundaries = copy.deepcopy(boundaries)
    
#     rhs = np.zeros(F.shape)#np.zeros(list(F.shape[:-2]) + [np.prod(F.shape[-2:])])
      
#     for key in boundaries.keys():
#         if len(boundaries[key].shape) > 1:
#             boundaries[key] = tf.squeeze(boundaries[key])
#         if len(boundaries[key].shape) == 1:
#             boundaries[key] = itertools.repeat(boundaries[key], F.shape[0])
    
#     if (h is not None) and isinstance(h, float):
#         h = itertools.repeat(h, F.shape[0])
      
#     for i in range(F.shape[0]):
#         if h is not None:
#             try:
#                 dx = next(h)
#             except:
#                 dx = h[i]
#             F[i] = -dx**2 * F[i]
      
#         if isinstance(boundaries['top'], Iterator):
#             top = next(boundaries['top'])
#         else:
#             top = boundaries['top'][i]
            
#         if isinstance(boundaries['bottom'], Iterator):
#             bottom = next(boundaries['bottom'])
#         else:
#             bottom = boundaries['bottom'][i]
            
#         if isinstance(boundaries['left'], Iterator):
#             left = next(boundaries['left'])
#         else:
#             left = boundaries['left'][i]
            
#         if isinstance(boundaries['right'], Iterator):
#             right = next(boundaries['right'])
#         else:
#             right = boundaries['right'][i]
        
      
#         rhs[i,...,0] = bottom
#         rhs[i,...,-1] = top
#         rhs[i,...,0,:] = left
#         rhs[i,...,-1,:] = right
#         rhs[i,...,1:-1,1:-1] = F[i,...,1:-1,1:-1]
      
#     return rhs.reshape(list(rhs.shape[:-2]) + [np.prod(rhs.shape[-2:])])

def multigrid_poisson_solve(rhses, boundaries, dx, dy = None, system_matrix = None, tol = 1e-10):
    '''
    Solves the Poisson equation for the given RHSes.
    
    rhses: tf.Tensor representing the RHS functions of the Poisson equation, defined across the last 2 dimensions
    boundaries: boundary conditions of the outputs; see poisson_RHS documentation
    dx: grid spacing of the outputs
    system_matrix: Sparse Poisson equation system matrix. Generated by pyamg.gallery.poisson
    tol: Tolerance of the multigrid solver.
    
    Outputs a tf.Tensor of identical shape to rhses.
    
    This function works on CPU.
    '''
    try:
        rhses = rhses.numpy()
    except:
        pass
    
    if rhses.shape[1] == 1:
        rhses = np.squeeze(rhses, axis = 1)
    
    solns = np.zeros([rhses.shape[0], 1] + [dim for dim in rhses.shape[1:]])
    rhs_vectors = poisson_RHS(rhses, boundaries, h = dx)

    if system_matrix == None:
        system_matrix = pyamg.gallery.poisson([dim-2 for dim in rhses.shape[1:]], format = 'csr')
    solver = pyamg.ruge_stuben_solver(system_matrix)

    interior_slice = [0, Ellipsis] + [slice(1,-1) for k in rhses.shape[1:]]
        
    for k in range(rhses.shape[0]):
        interior_slice[0] = k
        solns[interior_slice] = solver.solve(np.squeeze(rhs_vectors[k,...]), tol = tol).reshape([dim-2 for dim in rhses.shape[1:]], order = 'c')

    solns[...,:,-1] = boundaries['top']
    solns[...,:,0] = boundaries['bottom']
    solns[...,0,:] = boundaries['left']
    solns[...,-1,:] = boundaries['right']

    return solns
    



import tensorflow as tf
import numpy as np
import itertools
import opt_einsum as oe
import string
import copy

from .custom_blocks import ResampledConvolutionBlock, ResnetBlock, AveragePoolingBlock, Scaling
from .Model_With_Integral_Loss import Model_With_Integral_Loss_ABC
from ..layers import MergeWithAttention
from ..layers import Upsample
from ..layers import SpatialPyramidPool
#from ..layers import Scaling

def channels_first_rot_90(image,k=1):
    if len(image.shape) == 4:
        image = tf.transpose(image, (0,2,3,1))
    elif len(image.shape) == 3:
        image = tf.transpose(image, (0,2,1))
        image_was_rank3 = True
    elif len(image.shape) == 2:
        image = tf.expand_dims(image, axis = 2)
        image_was_rank3 = False
    else:
        raise ValueError('image must be a rank 2,3 or 4 Tensor')
    
    image = tf.image.rot90(image, k = k)
    
    if len(image.shape) == 4:
        return tf.transpose(image, (0,3,1,2))
    elif image_was_rank3:
        return tf.transpose(image, (0,2,1))
    else:
        return image[...,0]
    
class Homogeneous_Poisson_NN_Fluidnet_IA(Model_With_Integral_Loss_ABC):
    '''
    Takes a tf.Tensor containing the RHS and grid spacing information, and gives the corresponding solution to the Poisson problem with homogeneous BCs.
    '''
    def __init__(self, pooling_block_number = 6, post_dx_einsum_conv_block_number = 5, initial_kernel_size = 19, final_kernel_size = 3, resize_methods = None, data_format = 'channels_first', use_batchnorm = False, use_deconv_upsample = False, kernel_regularizer = None, bias_regularizer = None, use_scaling = True, scaling_args = {'downsampling_ratio_per_stage': 3, 'stages': 3, 'filters': 4, 'activation': tf.nn.leaky_relu, 'spp_levels': [[2,2],3,5]}, **kwargs):
        '''
        Init arguments:

        data_format: same as keras
        pooling_block_number: integer. controls the number of simultaneous pooling 'threads'. each pooling thread will apply average pooling with pool_size 2^k (k = [1,...,pooling_block_number]), do convoltuions and then upsample to original resolution. choose such that 2^(pooling_block_number) is close as possible to the minimum grid size your inputs will have.
        use_deconv_upsample: boolean. if set to true, pooling blocks will use transposed convolution to upsample. last 2 pooling layers will use bilinear and nearest neighbor unless overridden by resize_methods
        resize_methods: list of tf.image.ResizeMethod members. used to choose resizing algorithms for pooling block upsampling.
        use_batchnorm: boolean. determines if batch norm layers should be used. if set to true, supply __call__ argument training = False when performing inference.
        post_dx_einsum_conv_block_number: integer. no of convolution blocks after the result of the domain info processing is merged.
        initial_kernel_size: integer. initial kernel size of convolutions. most convolutions will have this kernel size.
        final_kernel_size: integer. the kernel size of the final convolutions will progressively move towards this number. keep it small to prevent output artefacting near image edges.
        kernel_regularizer: same as the corresponding tf.keras.layers.Conv2D argument
        bias_regularizer: same as the corresponding tf.keras.layers.Conv2D argument

        **kwargs: used to set self.integral_loss parameters for the Model_With_Integral_Loss_ABC abstract base class
        '''
        super().__init__(**kwargs)
        self.training = True
        self.pooling_block_number = pooling_block_number
        self.data_format = data_format
        self.pooling_block_kernel_sizes = 9 * np.ones((self.pooling_block_number), dtype = np.int32)
        self.pooling_block_kernel_sizes[-2:] = 3
        self.pooling_block_kernel_sizes = list(self.pooling_block_kernel_sizes)
        self.use_batchnorm = use_batchnorm

        final_dense_layer_units = 32
        
        if not resize_methods: #set resizing methods for pooling blocks if using bicubic upsampling
            try:
                self.resize_methods = [tf.compat.v1.image.ResizeMethod.BICUBIC for i in range(self.pooling_block_number-2)] + [tf.compat.v1.image.ResizeMethod.BILINEAR, tf.compat.v1.image.ResizeMethod.NEAREST_NEIGHBOR]
            except:
                self.resize_methods = [tf.image.ResizeMethod.BICUBIC for i in range(self.pooling_block_number-2)] + [tf.image.ResizeMethod.BILINEAR, tf.image.ResizeMethod.NEAREST_NEIGHBOR]
        else:
            self.resize_methods = resize_methods

        self.pooling_block_filters = 32
        
        self.conv_1 = tf.keras.layers.Conv2D(filters = 8, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        #self.batchnorm_1 = tf.keras.layers.BatchNormalization(axis = 1)
        
        self.conv_2 = tf.keras.layers.Conv2D(filters = self.pooling_block_filters, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        #self.batchnorm_2 = tf.keras.layers.BatchNormalization(axis = 1)
        
        self.conv_3 = tf.keras.layers.Conv2D(filters = self.pooling_block_filters, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

        if use_deconv_upsample:
            use_deconv_upsample = [True for k in range(self.pooling_block_number-2)] + [False for k in range(2)]
        else:
            use_deconv_upsample = [False for k in range(self.pooling_block_number)]
        self.pooling_blocks = [AveragePoolingBlock(2**(i+1), resize_method = self.resize_methods[i], data_format = data_format, kernel_size = int(self.pooling_block_kernel_sizes[i]), filters = self.pooling_block_filters, activation = tf.nn.leaky_relu, use_resnetblocks = True, use_deconv_upsample = use_deconv_upsample[i], kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer) for i in range(pooling_block_number)]
        
        self.merge = MergeWithAttention()
        
        self.dx_einsum_conv = tf.keras.layers.Conv2D(filters = final_dense_layer_units, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.dx_einsum_resnet = ResnetBlock(filters = final_dense_layer_units, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

        self.post_dx_einsum_conv_blocks = []
        for k in range(post_dx_einsum_conv_block_number):
            ksize = initial_kernel_size + (k*(final_kernel_size - initial_kernel_size)//(post_dx_einsum_conv_block_number-1))
            filters = final_dense_layer_units-((k*final_dense_layer_units)//(post_dx_einsum_conv_block_number))
            block = []
            block.append(tf.keras.layers.Conv2D(filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            block.append(ResnetBlock(filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            self.post_dx_einsum_conv_blocks.append(block)

        if use_batchnorm:
            self.batchnorm_layers = []
            if self.data_format == 'channels_first':
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = 1)
            else:
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = -1)
            for k in range(post_dx_einsum_conv_block_number):
                if self.data_format == 'channels_first':
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = 1))
                else:
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = -1))

        self.conv_last = tf.keras.layers.Conv2D(filters = 1, kernel_size = final_kernel_size, activation='linear', data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.resnet_last = ResnetBlock(filters = 1, kernel_size = final_kernel_size, activation='linear', data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        
        self.dx_dense_0 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dx_dense_1 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dx_dense_2 = tf.keras.layers.Dense(final_dense_layer_units, activation = 'linear')

        self.use_scaling = use_scaling
        if use_scaling:
            self.scaling = Scaling(data_format = self.data_format, kernel_size = final_kernel_size, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer, **scaling_args)
        
    def call(self, inp):
        '''
        input arguments:
        
        inp[0]: tf.Tensor of shape (batch_size, 1, nx, ny) or (batch_size, nx, ny, 1) based on self.data_format. this is the RHSes.
        inp[1]: tf.Tensor of shape (batch_size, 1). this must be the grid spacing information.
        '''
        ia = []
        self.dx = inp[1]
        if self.data_format == 'channels_first':#compute domain info
            domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[3]), tf.keras.backend.floatx())], axis = 0))
        else:
            domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[1]), tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx())], axis = 0))

        out = self.conv_1(inp[0]) #initial convolutions
        ia.append(copy.deepcopy(out))
        out = self.conv_2(out)
        ia.append(copy.deepcopy(out))
        for pb in self.pooling_blocks:
            ia.append(pb(out))
        ia.append(self.conv_3(out))
        out = self.merge([self.conv_3(out)] + [pb(out) for pb in self.pooling_blocks])#pooling 'threads'

        out = self.dx_einsum_conv(out)
        out = self.dx_einsum_resnet(out)
        
        out = tf.einsum('ijkl, ij -> ijkl',out, self.dx_dense_2(self.dx_dense_1(self.dx_dense_0(domain_info)))) #incorporate domain info

        for block_num, block in enumerate(self.post_dx_einsum_conv_blocks): #perform additional convolutions
            for layer in block:
                out = layer(out)
            if self.use_batchnorm:
                out = self.batchnorm_layers[block_num](out, training = self.training)

        if self.use_batchnorm:
            out = self.batchnorm_last(out, training = self.training)
        out = self.conv_last(out)
        out = self.resnet_last(out)

        if self.use_scaling:
            out = self.scaling([out, inp[0]])

        ia.append(out)
        return ia

    def __call__(self, inp, training = True):#overload __call__ to allow freezing batch norm parameters
        self.training = training
        return super().__call__(inp)



import tensorflow as tf
import itertools
import opt_einsum as oe
import numpy as np
from abc import ABC, abstractmethod
import sys
sys.path.append('..')

from ..layers import WeightedContractionLayer
from ..layers import Upsample
from .Model_With_Integral_Loss import Model_With_Integral_Loss_ABC
from .custom_blocks import *
                 
class Dirichlet_BC_NN(Model_With_Integral_Loss_ABC):
    '''
    This model takes a Dirichlet boundary condition imposed on a flat domain, the grid spacing and the desired output shape and outputs the solution to the Laplace equation with 1 Dirichlet BC (rest of the BCs are homogeneous)
    '''
    def __init__(self, data_format = 'channels_first', x_output_resolution = 256, dx_shape = [128, 32], post_conv1d_block_number = 2, post_dx_conv_block_number = 5, initial_kernel_size = 19, final_kernel_size = 3, use_batchnorm = False, kernel_regularizer = None, bias_regularizer = None, conv1d_layer_number = 3, conv1d_final_channels = 256, **kwargs):
        '''
        Init arguments:

        data_format: same as keras
        x_output_resolution: integer. no of gridpoints of the output in the direction perpendicular to the boundary. can be changed at runtime.
        dx_shape: list of integers. the final dense layer processing domain will have a no of units corresponding to the product of the members of this list. the output is then reshaped to match this.
        conv1d_layer_number: integer. no of 1d convolution layers at the start.
        conv1d_final_channels: integer. no of channels the final conv1d layer has.
        post_conv1d_block_number: integer. no of convolution blocks after the result of the initial 1d convs is reshaped to 2d.
        post_dx_conv_block_number: integer. no of convolution blocks after the result of the domain info processing is merged.
        initial_kernel_size: integer. initial kernel size of convolutions. most convolutions will have this kernel size.
        final_kernel_size: integer. the kernel size of the final convolutions will progressively move towards this number. keep it small to prevent output artefacting near image edges.
        use_batchnorm: boolean. determines if batch norm layers should be used. if set to true, supply __call__ argument training = False when performing inference.
        kernel_regularizer: same as the corresponding tf.keras.layers.Conv2D argument
        bias_regularizer: same as the corresponding tf.keras.layers.Conv2D argument

        **kwargs: used to set self.integral_loss parameters for the Model_With_Integral_Loss_ABC abstract base class
        '''
        super().__init__(**kwargs)
        self.x_output_resolution = x_output_resolution
        self.data_format = data_format
        self.dx_upsample_channels = 16
        self.post_dx_conv_block_number = post_dx_conv_block_number #no of conv blocks after merging doman information
        self.dx_shape = dx_shape
        self.use_batchnorm = use_batchnorm
        self.training = True
        
        self.conv1d_layers = []
        for k in range(conv1d_layer_number): #1d convs to convert 1d BC info to 2d data
            channels = conv1d_final_channels//(conv1d_layer_number - k)
            self.conv1d_layers.append(tf.keras.layers.Conv1D(filters = channels, kernel_size = initial_kernel_size, padding = 'same', activation = tf.nn.leaky_relu, data_format = self.data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))

        self.post_conv1d_blocks = [] #2d convs to increase channel size of the reshaped 1d conv results
        for k in range(post_conv1d_block_number):
            filters = self.dx_upsample_channels//(post_conv1d_block_number - k)
            block = []
            block.append(tf.keras.layers.Conv2D(filters = filters, kernel_size = initial_kernel_size, padding = 'same', activation = tf.nn.leaky_relu, data_format = data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            block.append(ResnetBlock(data_format = data_format, filters = filters, kernel_size = initial_kernel_size, activation = tf.nn.leaky_relu, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            self.post_conv1d_blocks.append(block)
                                                  
        self.dx_dense_0 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu) #dense layers that handle domain shape info
        self.dx_dense_1 = tf.keras.layers.Dense(256, activation = tf.nn.leaky_relu)
        self.dx_dense_2 = tf.keras.layers.Dense(tf.reduce_prod(dx_shape), activation = tf.nn.leaky_relu)
        self.dx_resampled_conv_blocks = [AveragePoolingBlock(pool_size = 2**i, data_format = data_format, filters = self.dx_upsample_channels, use_resnetblocks = True, use_deconv_upsample = True, activation = tf.nn.leaky_relu, kernel_size = 10-2*i) for i in range(4)]
        self.dx_upsample = Upsample([-1, -1], data_format = data_format)

        self.post_dx_conv_blocks = [] #2d convs after merging domain shape info and the result of self.post_conv1d_blocks
        for k in range(self.post_dx_conv_block_number):
            ksize = initial_kernel_size + (k*(final_kernel_size - initial_kernel_size)//(post_dx_conv_block_number-1))
            filters = self.dx_upsample_channels-((k*self.dx_upsample_channels)//(post_dx_conv_block_number))
            block = []
            block.append(tf.keras.layers.Conv2D(filters = filters, kernel_size = ksize, activation = tf.nn.leaky_relu, data_format = self.data_format, padding = 'same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            block.append(ResnetBlock(filters = filters, kernel_size = ksize , activation = tf.nn.leaky_relu, data_format = self.data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            self.post_dx_conv_blocks.append(block)

        if use_batchnorm: #batch norm layers to use if desired
            self.batchnorm_layers = []
            if self.data_format == 'channels_first':
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = 1)
            else:
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = -1)
            for k in range(post_dx_conv_block_number):
                if self.data_format == 'channels_first':
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = 1))
                else:
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = -1))
        
        self.conv_last = tf.keras.layers.Conv2D(filters = 1, kernel_size = final_kernel_size, activation = 'linear', padding = 'same', data_format = data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer) #final convolutions before output
        self.resnet_last = ResnetBlock(filters = 1, kernel_size = final_kernel_size, activation = 'linear', data_format = data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

    def call(self, inp):
       '''
       input arguments:
       
       inp[0]: tf.Tensor of shape (batch_size, 1, y direction dimension) or (batch_size, y direction dimension, 1) based on self.data_format. this must be the BC information.
       inp[1]: tf.Tensor of shape (batch_size, 1). this must be the grid spacing information.
       inp[2]: tf.Tensor of shape (4,) or (). this determines the grid size in the direction perpendicular to the boundary. if not provided, instead the last self.x_output_resolution will be used.
       '''
       self.dx = inp[1] 
       try:
           if inp[2].shape[0] == 1: #set output shape
               self.x_output_resolution = int(inp[2])
           elif self.data_format == 'channels_first':
               self.x_output_resolution = int(inp[2][2])
           else:
               self.x_output_resolution = int(inp[2][1])
       except:
           pass

       out = self.conv1d_layers[0](inp[0]) #1d convs
       for layer in self.conv1d_layers[1:]:
           out = layer(out)

       if self.data_format == 'channels_first': 
           out = tf.expand_dims(out, axis = 1) #convert 1d conv results into 2d
           domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(self.x_output_resolution, tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx())], axis = 0)) #calculate domain info
           input_length = inp[0].shape[2] #get mesh size in the direction of the provided BC
       else:
           out = tf.expand_dims(out, axis = -1)
           domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(self.x_output_resolution, tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[1]), tf.keras.backend.floatx())], axis = 0))
           input_length = inp[0].shape[1]

       for block in self.post_conv1d_blocks: #2d convs to increase channel no of the 1d conv results
           for layer in block:
               out = layer(out)

       dx_res = self.dx_dense_2(self.dx_dense_1(self.dx_dense_0(domain_info))) #process domain info
       if self.data_format == 'channels_first':
           dx_res = tf.keras.backend.reshape(dx_res, (-1, 1, self.dx_shape[0], self.dx_shape[1]))
       else:
           dx_res = tf.keras.backend.reshape(dx_res, (-1, self.dx_shape[0], self.dx_shape[1], 1))
       dx_upsample_result = self.dx_resampled_conv_blocks[0](dx_res)
       for layer in self.dx_resampled_conv_blocks[1:]:
           dx_upsample_result += layer(dx_res)

       try:#merge domain info
           out = self.dx_upsample([out, [self.x_output_resolution, input_length]]) + self.dx_upsample([dx_upsample_result, [self.x_output_resolution, input_length]])
       except:
           out = self.dx_upsample([out, [64, input_length]]) + self.dx_upsample([dx_upsample_result, [64, input_length]])

       for block_num, block in enumerate(self.post_dx_conv_blocks):#final convolutions to bring channel no to 1
           for layer in block:
               out = layer(out)
           if self.use_batchnorm: #apply batch norm if needed
               out = self.batchnorm_layers[block_num](out, training = self.training)

       if self.use_batchnorm:
           out = self.batchnorm_last(out, training = self.training)

       out = self.conv_last(out)
       out = self.resnet_last(out)
       
       return out

    def __call__(self, inp, training = True):#overload __call__ to allow freezing batch norm parameters
        self.training = training
        return super().__call__(inp)

            
        
        



import tensorflow as tf
import numpy as np

class Scaling(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.scaling_weight = self.add_weight(name = 'scaling_weight', shape = (), initializer=tf.initializers.ones, trainable = True)

    def call(self, inp):
        return inp * self.scaling_weight



import tensorflow as tf
import itertools
import opt_einsum as oe
import numpy as np
from abc import ABC, abstractmethod

class Model_With_Integral_Loss_ABC(ABC, tf.keras.models.Model):
    '''
    Abstract class to implement integral loss.
    '''
    def __init__(self, mae_component_weight = 0.0, mse_component_weight = 0.0, n_quadpts= 20, Lp_norm_power = 2, **kwargs):
        super().__init__( **kwargs)
        self.mae_component_weight = mae_component_weight
        self.mse_component_weight = mse_component_weight
        self.n_quadpts = n_quadpts
        self.p = Lp_norm_power

    def integral_loss(self, y_true, y_pred):
        try: #try except block to handle keras model init shenanigans
            if self.data_format == 'channels_first':
                c = 0.5* tf.concat([self.dx * (int(y_true.shape[-2])-1), self.dx * (int(y_true.shape[-1])-1)],1)
                coords = np.array(np.meshgrid(np.linspace(-1, 1, y_true.shape[-2]),np.linspace(-1, 1, y_true.shape[-1]),indexing = 'ij'), dtype = tf.keras.backend.floatx()).transpose((1,2,0)) #coordinates of each grid pt in the domain
            else:
                c = 0.5 * tf.concat([self.dx * int(y_true.shape[-3]), self.dx * int(y_true.shape[-2])],1)
                coords = np.array(np.meshgrid(np.linspace(-1, 1, y_true.shape[-3]),np.linspace(-1, 1, y_true.shape[-2]),indexing = 'ij'), dtype = tf.keras.backend.floatx()).transpose((1,2,0)) #coordinates of each grid pt in the domain
        except:
            return 0.0*(y_true - y_pred)
        image_coords = [coords[:,0,0], coords[1,:,1]] #x and y coordinates separately
        quadrature_x, quadrature_w = tuple([x.astype(tf.keras.backend.floatx()) for x in np.polynomial.legendre.leggauss(self.n_quadpts)])
        #quadrature_x, quadrature_w = tuple([np.polynomial.legendre.leggauss(self.n_quadpts)[i].astype(np.float64) for i in range(2)])

        quadpts = tf.constant(np.array(np.meshgrid(quadrature_x,quadrature_x,indexing = 'ij')).transpose((1,2,0)),dtype = tf.keras.backend.floatx())
        #quadweights = tf.reduce_prod(c)*tf.tensordot(tf.squeeze(quadrature_w),tf.squeeze(quadrature_w),axes = 0)
        indices = [[],[]] #indices between each quadrature point lies - indices[0] is in x-dir and indices[1] is in the y-dir
        quad_coords = [quadpts[:,0,0], quadpts[1,:,1]] #x and y coordinates of each quad pt respectively
        #find the indices of coords between which every quad. pt. lies
        for i in range(len(indices)):
            j=0
            #does not work if more than 2 quad pts fall within 1 cell - fix later
            while len(indices[i]) < quadpts.shape[0] and j<image_coords[i].shape[0]:
                try:
                    if abs(float(quad_coords[i][len(indices[i])] - image_coords[i][j])) == float(min(abs(quad_coords[i][len(indices[i])] - image_coords[i][j-1]), abs(quad_coords[i][len(indices[i])] - image_coords[i][j]), abs(quad_coords[i][len(indices[i])] - image_coords[i][j+1]))):
                        if quad_coords[i][len(indices[i])] - image_coords[i][j] < 0:
                            indices[i].append((j-1,j))
                        else:
                            indices[i].append((j,j+1))
                except:
                    if abs(float(quad_coords[i][len(indices[i])] - image_coords[i][j])) == float(min(abs(quad_coords[i][len(indices[i])] - image_coords[i][j-1]), abs(quad_coords[i][len(indices[i])] - image_coords[i][j]))):
                        indices[i].append((j-1,j))
                j+=1
        
        index_combinations = tf.Variable(tf.zeros((quadpts.shape[0], quadpts.shape[1], 4 , 2), dtype = tf.int32), trainable = False, dtype = tf.int32) #array storing the 4 index combinations on the original grid which surround each quad. pt.
        corners = tf.Variable(tf.zeros((quadpts.shape[0], quadpts.shape[1], 2 , 2), dtype = np.int32), dtype = tf.int32, trainable = False) #array storing the lower left corner and the upper right corner of each box stored in index_combinations. effectively this will contain [[xmin,ymin],[xmax,ymax]] for the rectangle around each quad pt.
        s=tf.constant(indices)

        for i in range(self.n_quadpts):
            for j in range(self.n_quadpts):
                index_combinations[i,j,:,:].assign(np.array(list(itertools.product(np.array(s)[0,i,:],np.array(s)[1,j,:]))))
        for i in range(self.n_quadpts):
            for j in range(self.n_quadpts):
                corners[i,j,:,:].assign(tf.stack([s[0,i,:],s[1,j,:]]))
        corners = tf.transpose(corners,(0,1,3,2))
        corner_coords = tf.gather_nd(tf.transpose(coords,(0,1,2)),corners)
        
        #compute the coefficients [b_11,b_12,b_21,b_22]
        #steps:
        #1. compute transpose(invert(array([[1,xmin,ymin,xmin*ymin],[1,xmin,ymax,xmin*ymax],[1,xmax,ymin,xmax*ymin],[1,xmax,ymax,xmax*ymax]]))) for the rectangle around each quad pt.
        #2. compute array([1,x_quadpt, y_quadpt, x_quadpt*y_quadpt]) for each quadpt
        #3. multiply the result of 1 and 2 for each quad pt.
        interpolation_matrix = tf.Variable(tf.ones((self.n_quadpts,self.n_quadpts,4,4), dtype = tf.keras.backend.floatx()), dtype = tf.keras.backend.floatx())
        interpolation_matrix[:,:,0:2,1].assign(oe.contract('ijk,ij->ijk',interpolation_matrix[:,:,0:2,1],corner_coords[:,:,0,0], backend = 'tensorflow'))
        interpolation_matrix[:,:,2:,1].assign(oe.contract('ijk,ij->ijk',interpolation_matrix[:,:,2:,1],corner_coords[:,:,1,0], backend = 'tensorflow'))
        interpolation_matrix[:,:,0::2,2].assign(oe.contract('ijk,ij->ijk',interpolation_matrix[:,:,0::2,2],corner_coords[:,:,0,1], backend = 'tensorflow'))
        interpolation_matrix[:,:,1::2,2].assign(oe.contract('ijk,ij->ijk',interpolation_matrix[:,:,1::2,2], corner_coords[:,:,1,1], backend = 'tensorflow'))
        interpolation_matrix[:,:,:,3].assign(oe.contract('...,...,...->...',interpolation_matrix[:,:,:,3],interpolation_matrix[:,:,:,1], interpolation_matrix[:,:,:,2], backend = 'tensorflow'))
        interpolation_matrix = tf.transpose(tf.linalg.inv(interpolation_matrix), (0,1,3,2))
        q = tf.Variable(tf.ones((self.n_quadpts,self.n_quadpts,4), dtype = tf.keras.backend.floatx()), dtype = tf.keras.backend.floatx())
        q[:,:,1].assign(tf.transpose(quadpts[:,:,0]))
        q[:,:,2].assign(tf.transpose(quadpts[:,:,1]))
        q[:,:,3].assign(tf.multiply(q[:,:,1],q[:,:,2]))
        
        b = oe.contract('ijkl, ijl->ijk', interpolation_matrix, q, backend = 'tensorflow')

        quadweights = oe.contract('i,j,k->ijk',tf.squeeze(quadrature_w),tf.squeeze(quadrature_w), tf.reduce_prod(c,axis = 1), backend = 'tensorflow')
        if self.data_format == 'channels_first':
            interp_pts = tf.squeeze(tf.gather_nd(tf.transpose(y_true - y_pred, (2,3,1,0)), index_combinations), axis = 3)
        else:
            interp_pts = tf.squeeze(tf.gather_nd(tf.transpose(y_true - y_pred, (1,2,3,0)), index_combinations), axis = 3)
        
        values_at_quad_pts = oe.contract('ijkl, ijk->ijl', interp_pts, b, backend = 'tensorflow')

        inverse_domain_area = 1/tf.reduce_prod(2*c, axis = 1)
    
        loss = tf.reduce_mean(oe.contract('i,i...->i...',inverse_domain_area,tf.reduce_sum(tf.multiply(quadweights, values_at_quad_pts**self.p), axis = (0,1)), backend = 'tensorflow')**(1/self.p))
        
        if self.mae_component_weight != 0.0:
            loss = loss + self.mae_component_weight * tf.reduce_mean(tf.keras.losses.mean_absolute_error(y_true, y_pred))
        if self.mse_component_weight != 0.0:
            loss = loss + self.mse_component_weight * tf.reduce_mean(tf.keras.losses.mean_squared_error(y_true, y_pred))
        return loss



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="poisson_CNN",
    version="0.2.0",
    author="Ali Girayhan Ozbay",
    author_email="aligirayhan.ozbay14@imperial.ac.uk",
    description="A convolutional neural network based Poisson solver",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aligirayhanozbay/poisson_CNN",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering"
    ],
    install_requires=[
        "numpy",
        "opt-einsum",
        "tensorflow-gpu>=2.0.0",
        "pyamg"
    ],
    python_requires='>=3.6',
)



import tensorflow as tf
import numpy as np
import opt_einsum as oe
import itertools


from .Homogeneous_Poisson_NN import *
from .Dirichlet_BC_NN import *
from ..dataset.generators.numerical import set_max_magnitude_in_batch as smmib

#helper methods to do rotation/reflection operations with data_format = 'channels_first'
def channels_first_rot90(image,k=1):
    if len(image.shape) == 4:
        image = tf.transpose(image, (0,2,3,1))
    elif len(image.shape) == 3:
        image = tf.transpose(image, (1,2,0))
        image_was_rank3 = True
    elif len(image.shape) == 2:
        image = tf.expand_dims(image, axis = 2)
        image_was_rank3 = False
    else:
        raise ValueError('image must be a rank 2,3 or 4 Tensor')
    
    image = tf.image.rot90(image, k = k)
    
    if len(image.shape) == 4:
        return tf.transpose(image, (0,3,1,2))
    elif image_was_rank3:
        return tf.transpose(image, (2,0,1))
    else:
        return image[...,0]
    
def channels_first_flip_up_down(image):
    if len(image.shape) == 4:
        image = tf.transpose(image, (0,2,3,1))
    elif len(image.shape) == 3:
        image = tf.transpose(image, (1,2,0))
        image_was_rank3 = True
    elif len(image.shape) == 2:
        image = tf.expand_dims(image, axis = 2)
        image_was_rank3 = False
    else:
        raise ValueError('image must be a rank 2,3 or 4 Tensor')
    
    image = tf.image.flip_left_right(image) #tensorflow seems to have the opposite alignment...
    
    if len(image.shape) == 4:
        return tf.transpose(image, (0,3,1,2))
    elif image_was_rank3:
        return tf.transpose(image, (2,0,1))
    else:
        return image[...,0]
    
def channels_first_flip_left_right(image):
    if len(image.shape) == 4:
        image = tf.transpose(image, (0,2,3,1))
    elif len(image.shape) == 3:
        image = tf.transpose(image, (1,2,0))
        image_was_rank3 = True
    elif len(image.shape) == 2:
        image = tf.expand_dims(image, axis = 2)
        image_was_rank3 = False
    else:
        raise ValueError('image must be a rank 2,3 or 4 Tensor')
    
    image = tf.image.flip_up_down(image) #tensorflow seems to have the opposite alignment...
    
    if len(image.shape) == 4:
        return tf.transpose(image, (0,3,1,2))
    elif image_was_rank3:
        return tf.transpose(image, (2,0,1))
    else:
        return image[...,0]

class Poisson_CNN(Model_With_Integral_Loss_ABC):
    '''
    A CNN model that predicts the solution of a Poisson problem with Dirichlet BCs
    '''
    def __init__(self, bc_nn_parameters = {}, homogeneous_poisson_nn_parameters = {}, bc_nn_weights = None,  homogeneous_poisson_nn_weights = None, bc_nn_trainable = True, homogeneous_poisson_nn_trainable = True, data_format = 'channels_first', **kwargs):
        '''
        Init arguments:

        bc_nn_parameters: Dict containing arguments to be supplied to Dirichlet_BC_NN as Dirichlet_BC_NN(**bc_nn_parameters)
        homogeneous_poisson_nn_parameters: Dict containing arguments to be supplied to Homogeneous_Poisson_NN
        bc_nn_weights: String or None. If a string, it must be the (relative) filepath to a HDF5 file contanining the weights for Dirichlet_BC_NN. No weights will be loaded if left as None.
        homogeneous_poisson_nn_weights: String or None. If a string, it must be the (relative) filepath to a HDF5 file contanining the weights for Homogeneous_Poisson_NN. No weights will be loaded if left as None.
        bc_nn_trainable: If set to True, BC NN weights will not be frozen during training.
        homogeneous_poisson_nn_trainable: If set to True, Homogeneous_Poisson_NN weighrs will not be frozen during training.
        data_format: Same as tf.keras.layers.Conv2D
        
        **kwargs: used to set self.integral_loss parameters for the Model_With_Integral_Loss_ABC abstract base class
        '''
        super().__init__(**kwargs)

        #choose image reflection/rotation functions
        self.data_format = data_format
        if self.data_format == 'channels_first':
            self.rotation_method = channels_first_rot90
            self.fliplr_method = channels_first_flip_left_right
            self.flipud_method = channels_first_flip_up_down
            self.stacking_dim = 1
        else:
            self.rotation_method = tf.image.rot90
            self.fliplr_method = tf.image.flip_up_down
            self.flipud_method = tf.image.flip_left_right
            self.stacking_dim = 3
        if 'data_format' not in bc_nn_parameters.keys():
            bc_nn_parameters['data_format'] = self.data_format
        if 'data_format' not in homogeneous_poisson_nn_parameters.keys():
            homogeneous_poisson_nn_parameters['data_format'] = self.data_format

        #instantiate Dirichlet_BC_NN, initialize weights, load weight file
        self.bc_nn = Dirichlet_BC_NN(**bc_nn_parameters)
        self.bc_nn((tf.random.uniform((10,1,74), dtype = tf.keras.backend.floatx()), tf.random.uniform((10,1), dtype = tf.keras.backend.floatx()))) #initialize weights
        try:
            self.bc_nn.load_weights(bc_nn_weights)
        except:
            pass
        for layer in self.bc_nn.layers:
            layer.trainable = bc_nn_trainable

        #instantiate Homogeneous_Poisson_NN, initialize weights, load weight file
        self.hpnn = Homogeneous_Poisson_NN_Fluidnet(**homogeneous_poisson_nn_parameters)
        self.hpnn((tf.random.uniform((10,1,74,83), dtype = tf.keras.backend.floatx()), tf.random.uniform((10,1), dtype = tf.keras.backend.floatx())))
        try:
            self.hpnn.load_weights(homogeneous_poisson_nn_weights)
        except:
            pass
        for layer in self.hpnn.layers:
            layer.trainable = homogeneous_poisson_nn_trainable
        
        
    def call(self, inp):# rhs, boundaries, dx):
        '''
        Call arguments:

        inp[0]: tf.Tensor of shape (batch_size, 1, nx, ny) or (batch_size, nx, ny, 1) based on self.data_format. this is the RHSes.
        inp[1]: tf.Tensor of shape (batch_size, 1, ny) or (batch_size, ny, 1) based on self.data_format. this is the left Dirichlet BC.
        inp[2]: tf.Tensor of shape (batch_size, 1, nx) or (batch_size, nx, 1) based on self.data_format. this is the top Dirichlet BC.
        inp[3]: tf.Tensor of shape (batch_size, 1, ny) or (batch_size, ny, 1) based on self.data_format. this is the right Dirichlet BC.
        inp[4]: tf.Tensor of shape (batch_size, 1, nx) or (batch_size, nx, 1) based on self.data_format. this is the bottom Dirichlet BC.
        inp[5]: tf.Tensor of shape (batch_size, 1). this must be the grid spacing information.
        '''

        #scale all inputs to have 1.0 peak magnitudes
        rhs, rhs_scaling_factors = smmib(inp[0], 1.0, return_scaling_factors = True)
        left_boundary, left_boundary_scaling_factors = smmib(inp[1], 1.0, return_scaling_factors = True)
        top_boundary, top_boundary_scaling_factors = smmib(inp[2], 1.0, return_scaling_factors = True)
        right_boundary, right_boundary_scaling_factors = smmib(inp[3], 1.0, return_scaling_factors = True)
        bottom_boundary, bottom_boundary_scaling_factors = smmib(inp[4], 1.0, return_scaling_factors = True)
        scaling_factors = tf.stack([left_boundary_scaling_factors, right_boundary_scaling_factors, top_boundary_scaling_factors, bottom_boundary_scaling_factors, rhs_scaling_factors], axis = 1)
        
        self.dx = inp[5]
        if len(self.dx.shape) == 1:
            self.dx = tf.expand_dims(self.dx,axis = 1)

        if self.data_format == 'channels_first':
            self.bc_nn.x_output_resolution = rhs.shape[2]#predict for left and right BCs by supplying them to Dirichlet_BC_NN as a single batch
            tmp = self.bc_nn([tf.concat([left_boundary, right_boundary], axis = 0), tf.tile(self.dx, [2,1])])
            left_boundary = tmp[0:rhs.shape[0],0,...]
            right_boundary = self.flipud_method(self.rotation_method(tmp[rhs.shape[0]:,0,...], k = 2)) #rotate and reflect right boundary method

            self.bc_nn.x_output_resolution = rhs.shape[3]#predict for top and bottom BCs by supplying them to Dirichlet_BC_NN as a single batch
            tmp = self.bc_nn([tf.concat([bottom_boundary, top_boundary], axis = 0), tf.tile(self.dx, [2,1])])
            bottom_boundary = self.fliplr_method(self.rotation_method(tmp[0:rhs.shape[0],0,...], k = 1)) #rotate and reflect top boundary prediction
            top_boundary = self.rotation_method(tmp[rhs.shape[0]:,0,...], k = 3) #rotate bottom boundary prediction

            rhs = tf.squeeze(self.hpnn([rhs, self.dx]), axis = self.stacking_dim) #predict for the homogeneous Poisson problem

            solutions = tf.stack([left_boundary, right_boundary, top_boundary, bottom_boundary, rhs], axis = 1) 
            solutions = tf.einsum('ij...,ij->ij...', solutions, 1/scaling_factors) #revert each prediction to original scale
            
        else: #same procedures but for channels_last
            self.bc_nn.x_output_resolution = rhs.shape[2]
            tmp = self.bc_nn([tf.concat([left_boundary, right_boundary], axis = 0), tf.tile(self.dx, [2,1])])
            left_boundary = tmp[0:rhs.shape[0],...,0]
            right_boundary = self.flipud_method(self.rotation_method(tmp[rhs.shape[0]:,...,0], k = 2))

            self.bc_nn.x_output_resolution = rhs.shape[3]
            tmp = self.bc_nn([tf.concat([top_boundary, bottom_boundary], axis = 0), tf.tile(self.dx, [2,1])])
            top_boundary = self.fliplr_method(self.rotation_method(tmp[0:rhs.shape[0],...,0], k = 1))
            bottom_boundary = self.rotation_method(tmp[rhs.shape[0]:,...,0], k = 3)

            rhs = tf.squeeze(self.hpnn([rhs, self.dx]))

            solutions = tf.stack([left_boundary, right_boundary, top_boundary, bottom_boundary, rhs], axis = -1)
            solutions = tf.einsum('i...j,ij->i...j', solutions, 1/scaling_factors)
        
        return tf.reduce_sum(solutions, axis = self.stacking_dim, keepdims = True)



import tensorflow as tf
import itertools
import numpy as np
import multiprocessing
from collections.abc import Iterable

from ..solvers import *


@tf.function
def image_resize(image, newshape, data_format = 'channels_first', resize_method = tf.image.ResizeMethod.BICUBIC):
    '''
    Helper function to resize images on the CPU in parallel
    '''
    imagedims = len(image.shape)
    if data_format == 'channels_first':
        if len(image.shape) == 4:
            image = tf.transpose(image, (0,2,3,1))
        elif len(image.shape) == 3:
            image = tf.expand_dims(image, axis = 3)
        elif len(image.shape) == 2:
            image = tf.expand_dims(tf.expand_dims(image, axis = 2), axis = 0)
    if isinstance(newshape, list) or isinstance(newshape, tuple) or len(newshape.shape) == 1:
        newshape = tf.tile(tf.constant([newshape]), tf.constant([image.shape[0],1]))
        
    out = tf.cast(tf.map_fn(lambda x: tf.compat.v1.image.resize_images(x[0],x[1],method=resize_method,align_corners=True), (image, newshape), parallel_iterations=multiprocessing.cpu_count(), dtype = tf.float32), image.dtype)
    
    if data_format == 'channels_first':
        if imagedims == 4:
            out = tf.transpose(out, (0,3,1,2))
        elif imagedims == 3:
            out = out[...,0]
        else:
            out = out[0,...,0]
            
    return out

@tf.function
def set_max_magnitude(arr, max_magnitude = None, return_scaling_factors = False):
    '''
    Helper method to set the max magnitude in an array
    '''
    if max_magnitude == None:
        max_magnitude = arr[1]
        arr = arr[0]
        
    scaling_factor = max_magnitude/tf.reduce_max(tf.abs(arr))
    if return_scaling_factors:
        return arr * scaling_factor, scaling_factor
    else:
        return arr * scaling_factor
    
@tf.function
def set_max_magnitude_in_batch(arr, max_magnitude, return_scaling_factors = False):
    '''
    Set max magnitudes within each batch in parallel
    '''
    if isinstance(max_magnitude, float):
        #max_magnitude = tf.cast(tf.constant([max_magnitude for k in range(arr.shape[0])]), arr.dtype)
        max_magnitude = tf.cast(tf.keras.backend.tile(tf.constant([max_magnitude]), [tf.shape(arr)[0]]), arr.dtype)

    if return_scaling_factors:
        return_dtype = (arr.dtype, arr.dtype)
    else:
        return_dtype = arr.dtype
    
    return tf.map_fn(lambda x: set_max_magnitude(x, return_scaling_factors = return_scaling_factors), (arr, max_magnitude), dtype = return_dtype, parallel_iterations = multiprocessing.cpu_count())

def generate_random_RHS(batch_size, n_outputpts, smoothness = None, resize_method = tf.image.ResizeMethod.BICUBIC, max_magnitude = np.inf):
    
    '''
    This function generates random smooth RHS 'functions' defined pointwise by creating a random number field the shape of which is defined by n_controlpts and then super-sampling it using tf.image.resize_images. Results are stacked across the 1st dimension.
    
    batch_size: no. of random RHSes to generate
    n_outputpts: Iterable. No. of gridpoints in each direction of the output 
    smoothness: None or Iterable or int. No. of control pts of the spline along each dimension. Smaller values lead to 'smoother' results. If None, a value between 5 and n_outputpts[k]//1.5 is chosen randomly.
    resize_method Supersampling method for tf.image.resize_images. Use bicubic for smooth RHSes - bilinear or nearest neighbor NOT recommended
    max_random_magnitude: If the maximum absolute value of a generated random RHS exceeds this value, the entire thing is rescaled so the max abs. val. becomes this value
    '''
    n_controlpts = smoothness
    if n_controlpts == None:
        n_controlpts = [np.random.randint(5,int(n_outputpts[k]//1.5)) for k in range(len(n_outputpts))]
    elif isinstance(n_controlpts, int):
        try:
            n_controlpts = tf.ones(n_outputpts.shape, dtype = tf.int32)*n_controlpts
        except:
            n_controlpts = tf.ones(len(n_outputpts), dtype = tf.int32)*n_controlpts
    rhs = 2*tf.random.uniform([batch_size, 1] + list(n_controlpts), dtype = tf.keras.backend.floatx())-1
    rhs = image_resize(rhs, n_outputpts, resize_method = resize_method)
    rhs = rhs
    if max_magnitude != np.inf:
        rhs = set_max_magnitude_in_batch(rhs, max_magnitude)

    return rhs

def generate_random_boundaries(n_outputpts, batch_size = 1, max_magnitude = {'left':np.inf, 'top':np.inf, 'right':np.inf, 'bottom':np.inf}, smoothness = None, nonzero_boundaries = ['left', 'right', 'bottom', 'top'], return_with_expanded_dims = False, data_format = 'channels_first'):
    '''
    This function generates random smooth BC 'functions' defined pointwise by creating a random number field the size of which is determined by 'smoothness' and then super-sampling it using tf.image.resize_images. Results are stacked across the 1st dimension.
    
    batch_size: no. of random RHSes to generate
    n_outputpts: Iterable or int. No. of gridpoints in each direction of the output 
    resize_method: Supersampling method for tf.image.resize_images. Use bicubic for smooth boundaries - bilinear or nearest neighbor NOT recommended
    max_random_magnitude: Dict with entries 'left', 'right', 'top', 'bottom'. If the maximum absolute value of a generated random BC exceeds this value, it's rescaled so the max abs. val. becomes this value.
    nonzero_boundaries: Boundaries not included in this list will be all 0s. Entries same as max_random_magnitude.
    return_with_expanded_dims: If set to true, the returned arrays will have shape (batch_size,1,n) instead of (batch_size,n)
    data_format: If set to 'channels_first', the expanded dim will be the 2nd dimension. Else, it'll be the 3rd.
    
    Outputs a dict containing entries 'left', 'right', 'bottom' and 'top' with values as tf.Tensors of shape (batch_size, 1, n) or (batch_size, n) depending on whether return_with_expanded_dims is True, where n is the entry of n_outputpts corresponding to the boundary in question
    '''
    
    boundary_lengths = {'left' : n_outputpts[1], 'right' : n_outputpts[1], 'top' : n_outputpts[0], 'bottom' : n_outputpts[0]}
    if isinstance(smoothness, int):
        smoothness = {'left' : smoothness, 'right' : smoothness, 'top' : smoothness, 'bottom' : smoothness}
    elif smoothness == None:
        smoothness = {'left' : np.random.randint(5,int(n_outputpts[1]//1.5)), 'right' : np.random.randint(5,int(n_outputpts[1]//1.5)), 'top' : np.random.randint(5,int(n_outputpts[0]//1.5)), 'bottom' : np.random.randint(5,int(n_outputpts[0]//1.5))}
    boundaries = {}
    for boundary in smoothness.keys():
        if boundary in nonzero_boundaries:
            boundaries[boundary] = image_resize(2*tf.random.uniform((batch_size,smoothness[boundary]), dtype = tf.keras.backend.floatx())-1, [batch_size, boundary_lengths[boundary]])

            if max_magnitude[boundary] != np.inf:
                boundaries[boundary] = set_max_magnitude_in_batch(boundaries[boundary], max_magnitude[boundary])

        else:
            boundaries[boundary] = tf.zeros((batch_size, boundary_lengths[boundary]), dtype = tf.keras.backend.floatx())
        if return_with_expanded_dims:
            if data_format == 'channels_first':
                boundaries[boundary] = tf.expand_dims(boundaries[boundary], axis = 1)
            else:
                boundaries[boundary] = tf.expand_dims(boundaries[boundary], axis = 2)
    return boundaries

def numerical_dataset(batch_size = 1, output_shape = 'random', dx = 'random', boundaries = 'random', rhses = 'random', rhs_smoothness = None, boundary_smoothness = None, rhs_max_magnitude = 1.0, boundary_max_magnitude = {'left':1.0, 'top':1.0, 'right':1.0, 'bottom': 1.0}, nonzero_boundaries = ['left', 'right', 'bottom', 'top'], solver_method = 'multigrid', return_rhs = True, return_boundaries = False, return_dx = False, return_shape = False, random_output_shape_range = [[64,85],[64,85]], random_dx_range = [0.005,0.05]):
    '''
    Generates Poisson equation RHS-solution pairs with 'random' RHS functions.

    batch_size: No of problem-soln pairs to generate
    output_shape: shape of the outputs
    dx: Grid spacing. Can be 'random' for random grid spacings, a floating point value or an array of size (batch_size).
    boundaries: Supply the boundaries as a dict containing the entries 'top', 'bottom', 'right' and 'left' or set as 'random' for random BCs.
    rhses: Supply the RHSes as a tensor of shape (batch_size, 1, nx, ny) or set as 'random'
    smoothness: Int. Determines the resolution of the coarse grid.
    rhs_max_magnitude: Float. Max magnitude for generated rhses
    boundary_max_magnitude: Dict containing the entries 'top', 'bottom', 'right' and 'left'. Sets max magnitude for the corresponding boundaries.
    nonzero_boundaries: See generate_random_boundaries
    solver_method: 'multigrid' for PyAMG multigrid solver, 'cholesky' for Cholesky decomposition solver (on GPU) or supply a callable taking arguments (rhses, boundaries, dx)
    return_rhs: If set to True, the RHSes will be appended to the output
    return_boundaries: If set to True, the BCs will be appended to the output
    return_dx: If set to True, the grid spacing(s) will be appended to the output
    random_output_shape_range: List of lists, where each list contains [n_min, n_max] along the corresponding dim

    Outputs a tf.Tensor of shape (batch_size * smoothness_levels, n[0], n[1])
    '''

    if output_shape == 'random':
        output_shape = [np.random.randint(random_output_shape_range[k][0], high = random_output_shape_range[k][1]) for k in range(len(random_output_shape_range))]
    
    if rhses == 'random':
        rhses = generate_random_RHS(batch_size, output_shape, smoothness = rhs_smoothness, max_magnitude = rhs_max_magnitude)
    elif rhses == 'zero':
        rhses = np.zeros([batch_size,1] + list(output_shape))
    
    if boundaries == 'random':
        boundaries = generate_random_boundaries(output_shape, batch_size = batch_size, max_magnitude = boundary_max_magnitude, return_with_expanded_dims = True, nonzero_boundaries = nonzero_boundaries, smoothness = boundary_smoothness)
    elif boundaries == 'zero' or boundaries == 'homogeneous':
        boundaries = generate_random_boundaries(output_shape, batch_size = batch_size, return_with_expanded_dims = True, nonzero_boundaries = [])

    if dx == 'random':
        dx = tf.random.uniform((batch_size,1))*(random_dx_range[1] - random_dx_range[0]) + random_dx_range[0]
    elif isinstance(dx, float):
        dx = np.ones((batch_size,1))*dx
    
    if not callable(solver_method):
        if solver_method == 'cholesky':
            solver_method = cholesky_poisson_solve
        elif solver_method == 'multigrid':
            solver_method = multigrid_poisson_solve
        else:
            raise(ValueError('solver_method must be a function or one of cholesky or multigrid'))

    out = solver_method(rhses, boundaries,dx)
    inp = []
    if return_rhs:
        inp.append(rhses)
    if return_boundaries:
        inp.append(boundaries)
    if return_dx:
        inp.append(tf.cast(dx, tf.keras.backend.floatx()))
    if return_shape:
        inp.append(tf.cast(tf.constant(rhses.shape), tf.int32))
    
    if len(inp) > 0:
        return inp, tf.constant(out)
    else:
        return tf.constant(out)

class numerical_dataset_generator(tf.keras.utils.Sequence):
    def __init__(self, batch_size = 1, batches_per_epoch = 1, randomize_rhs_smoothness = False, rhs_random_smoothness_range = [5,20], randomize_boundary_smoothness = False, boundary_random_smoothness_range = {'left':[5,20], 'top':[5,20], 'right':[5,20], 'bottom': [5,20]}, randomize_rhs_max_magnitude = False, rhs_random_max_magnitude = 1.0, randomize_boundary_max_magnitudes = False, boundary_random_max_magnitudes = {'left':1.0, 'top':1.0, 'right':1.0, 'bottom': 1.0}, return_keras_style = True, exclude_zero_boundaries = False, **numerical_dataset_arguments):
        '''
        Creates a generator that can be used to generate infinitely many sets of Poisson eq. RHSes-BCs-solutions
        
        batch_size: Int. Batch size of outputs.
        batches_per_epoch: Int. No of batches to create in each training epoch.
        randomize_rhs_smoothness: Boolean. Set to True if it's desired to have random RHS smoothnesses.
        rhs_random_smoothness_range: List of 2 integers. First integer is the lower bound and 2nd integer is the upper bound of the random RHS smoothnesses. Ignored if randomize_rhs_smoothness is False.
        randomize_boundary_smoothness: Boolean. Set to True if it's desired to have random BC smoothnesses.
        boundary_random_smoothness_range: Dict containing entries 'left', 'top', 'right', 'bottom'. Each entry must be the same format as rhs_random_smoothness_range
        randomize_rhs_max_magnitude: Boolean. Set to True if it's desired to have random RHS magnitudes.
        rhs_random_max_magnitude: Float. Max value of the random max magnitude.
        randomize_boundary_max_magnitude: Boolean. Set to True if it's desired to have random BC magnitudes.
        boundary_random_max_magnitude: Dict containing entries 'left', 'top', 'right', 'bottom'. Each entry must be the same format as rhs_random_max_magnitude.
        numerical_dataset_arguments: Arguments to pass onto the function numerical_dataset.
        return_keras_style: If set to True, the values from the boundaries dict will be unpacked (in the order left-top-right-bottom) into members of the output list.
        exclude_zero_boundaries: If set to True, nonzero boundaries are not returned if return_keras_style is True.
        '''

        self.randomize_rhs_smoothness = randomize_rhs_smoothness
        if self.randomize_rhs_smoothness:
            self.rhs_random_smoothness_range = rhs_random_smoothness_range
        self.randomize_boundary_smoothness = randomize_boundary_smoothness
        if self.randomize_boundary_smoothness:
            self.boundary_random_smoothness_range = boundary_random_smoothness_range
        self.randomize_rhs_max_magnitude = randomize_rhs_max_magnitude
        if self.randomize_rhs_max_magnitude:
            self.rhs_random_max_magnitude = rhs_random_max_magnitude
        self.randomize_boundary_max_magnitudes = randomize_boundary_max_magnitudes
        if self.randomize_boundary_max_magnitudes:
            self.boundary_random_max_magnitudes = boundary_random_max_magnitudes

        self.exclude_zero_boundaries = exclude_zero_boundaries
        self.batch_size = batch_size
        self.batches_per_epoch = batches_per_epoch
        self.nda = numerical_dataset_arguments
        self.return_keras_style = return_keras_style

    def __len__(self):
        return self.batches_per_epoch

    def __getitem__(self, idx): #Generates a batch. Input idx is ignored but is necessary per Keras API.
        if self.randomize_rhs_smoothness:
            self.nda['rhs_smoothness'] = np.random.randint(self.rhs_random_smoothness_range[0], high = self.rhs_random_smoothness_range[1])
        if self.randomize_boundary_smoothness:
            self.nda['boundary_smoothness'] = dict(zip(list(self.boundary_random_smoothness_range.keys()),[np.random.randint(brsr[0], high = brsr[1]) for brsr in list(self.boundary_random_smoothness_range.values())]))
        if self.randomize_rhs_max_magnitude:
            self.nda['rhs_max_magnitude'] = np.random.rand() * self.rhs_random_max_magnitude
        if self.randomize_boundary_max_magnitudes:
            self.nda['boundary_max_magnitude'] = dict(zip(list(self.boundary_random_max_magnitudes.keys()),[np.random.rand()*brmm for brmm in list(self.boundary_random_max_magnitudes.values())]))
        inp, out = numerical_dataset(**self.nda, batch_size = self.batch_size)
        if self.return_keras_style:
            if ('return_boundaries' in self.nda.keys()):
                if ('return_rhs' not in self.nda.keys()) or self.nda['return_rhs']:
                    boundary_location = 1
                else:
                    boundary_location = 0
                if (not self.exclude_zero_boundaries) or ('nonzero_boundaries' not in self.nda.keys()):
                    boundary_list = [inp[boundary_location]['left'], inp[boundary_location]['top'], inp[boundary_location]['right'], inp[boundary_location]['bottom']]
                else:
                    boundary_list = [inp[boundary_location][key] for key in self.nda['nonzero_boundaries']]
                _ = inp.pop(boundary_location)
                inp[boundary_location:boundary_location] = boundary_list
        return inp, out
